# Copyright (c) 2013, Clemson University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the {organization} nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import datetime
import json
import os
import re
import subprocess
from abc import ABCMeta, abstractmethod, abstractproperty
from functools import wraps
from .hashdict import HashDict

multislash_rx = re.compile(r'//+')
isodate_rx = re.compile(r'(?P<year>\d{4})-?(?P<month>\d{2})-?(?P<day>\d{2})(?:\s*(?:T\s*)?(?P<hour>\d{2})(?::?(?P<minute>\d{2})(?::?(?P<second>\d{2}))?)?(?:[,.](?P<fraction>\d+))?(?:\s*(?P<tz>(?:Z|[+-](?P<tzhh>\d{2})(?::?(?P<tzmm>\d{2}))?)))?)')
tz_rx = re.compile(r'^(?P<tz>(?:Z|[+-](?P<tzhh>\d{2})(?::?(?P<tzmm>\d{2}))?))$')


def parse_isodate(datestr):
    """Parse a string that loosely fits ISO 8601 formatted date-time string
    """
    m = isodate_rx.search(datestr)
    assert m, 'unrecognized date format: ' + datestr
    year, month, day = m.group('year', 'month', 'day')
    hour, minute, second, fraction = m.group('hour', 'minute', 'second', 'fraction')
    tz, tzhh, tzmm = m.group('tz', 'tzhh', 'tzmm')
    dt = datetime.datetime(int(year), int(month), int(day), int(hour))
    if fraction is None:
        fraction = 0
    else:
        fraction = float('0.' + fraction)
    if minute is None:
        dt = dt.replace(minute=int(60 * fraction))
    else:
        dt = dt.replace(minute=int(minute))
        if second is None:
            dt = dt.replace(second=int(60 * fraction))
        else:
            dt = dt.replace(second=int(second), microsecond=int(1000000 * fraction))
    if tz is not None:
        if tz[0] == 'Z':
            offset = 0
        else:
            offset = datetime.timedelta(minutes=int(tzmm or 0), hours=int(tzhh))
            if tz[0] == '-':
                offset = -offset
        dt = dt.replace(tzinfo=UTCOffset(offset))
    return dt


class ABCMetaDocStringInheritor(ABCMeta):
    '''A variation on
    http://groups.google.com/group/comp.lang.python/msg/26f7b4fcb4d66c95
    by Paul McGuire
    '''
    def __new__(meta, name, bases, clsdict):
        if not('__doc__' in clsdict and clsdict['__doc__']):
            for mro_cls in (mro_cls for base in bases for mro_cls in base.mro()):
                doc = mro_cls.__doc__
                if doc:
                    clsdict['__doc__'] = doc
                    break
        for attr, attribute in clsdict.items():
            if not attribute.__doc__:
                for mro_cls in (
                    mro_cls for base in bases for mro_cls in base.mro()
                    if hasattr(mro_cls, attr)
                ):
                    doc = getattr(getattr(mro_cls, attr), '__doc__')
                    if doc:
                        attribute.__doc__ = doc
                        break
        return ABCMeta.__new__(meta, name, bases, clsdict)


class UnknownVCSType(Exception):
    pass


class RevisionPathException(Exception):
    def __init__(self, rev, path):
        super(RevisionPathException, self).__init__(rev, path)


class PathDoesNotExist(RevisionPathException):
    pass


class BadFileType(RevisionPathException):
    pass


class attrdict(dict):
    def __getattr__(self, name):
        return self.__getitem__(name)

    def __setattr__(self, name, value):
        if name.startswith('_'):
            dict.__setattr__(self, name, value)
        else:
            self.__setitem__(name, value)

    def __delattr__(self, name):
        self.__delitem__(name)


class CommitLogEntry(object):
    def __init__(self, rev, parents, date, author, message):
        self.rev = rev
        self.parents = parents
        self.date = date
        self.author = author
        self.message = message

    def __str__(self):
        return str(self.rev)

    def __repr__(self):
        return str('<%s.%s %s>' % (type(self).__module__, type(self).__name__, self.rev))

    @property
    def subject(self):
        return self.message.split('\n', 1)[0]

    def to_json(self):
        return json.dumps({
            'v': 1,
            'r': self.rev,
            'p': self.parents,
            'd': self.date.isoformat(),
            'a': self.author,
            'm': self.message,
        })

    @classmethod
    def from_json(cls, s):
        o = json.loads(s)
        if o.get('v') != 1:
            return None
        return cls(
            rev=o['r'],
            parents=o['p'],
            date=parse_isodate(o['d']),
            author=o['a'],
            message=o['m'],
        )


class CommitLogCache(HashDict):
    def __getitem__(self, key):
        value = HashDict.__getitem__(self, key)
        value = CommitLogEntry.from_json(value)
        if value:
            return value
        raise KeyError(key)

    def __setitem__(self, key, value):
        value = value.to_json()
        HashDict.__setitem__(self, key, value)


class FileChangeInfo(object):
    def __init__(self, path, status, copy=None):
        self.path = path
        self.status = status
        self.copy = copy


class BlameInfo(object):
    def __init__(self, rev, author, date, line):
        self.rev = rev
        self.author = author
        self.date = date
        self.line = line


class UTCOffset(datetime.tzinfo):
    ZERO = datetime.timedelta()

    def __init__(self, offset, name=None):
        if isinstance(offset, datetime.timedelta):
            self.offset = offset
        elif isinstance(offset, str):
            m = tz_rx.match(offset)
            assert m
            tz, tzhh, tzmm = m.group('tz', 'tzhh', 'tzmm')
            offset = datetime.timedelta(minutes=int(tzmm or 0), hours=int(tzhh))
            if tz[0] == '-':
                offset = -offset
            self.offset = offset
        else:
            self.offset = datetime.timedelta(minutes=offset)
        if name is not None:
            self.name = name
        elif self.offset < type(self).ZERO:
            self.name = '-%02d%02d' % divmod((-self.offset).seconds / 60, 60)
        else:
            self.name = '+%02d%02d' % divmod(self.offset.seconds / 60, 60)

    def utcoffset(self, dt):
        return self.offset

    def dst(self, dt):
        return type(self).ZERO

    def tzname(self, dt):
        return self.name


class VCSRepo(object):
    __metaclass__ = ABCMetaDocStringInheritor

    def __init__(self, path, encoding='utf-8'):
        """Open an existing repository"""
        self.path = path
        self.encoding = encoding

    @abstractproperty
    def private_path(self):
        """Get the path to a directory which can be used to store arbitrary data

        This directory should not conflict with any of the repository internals.
        The directory should be created if it does not already exist.

        """
        raise NotImplementedError

    @property
    def _commit_cache(self):
        try:
            return self._commit_cache_v
        except AttributeError:
            commit_cache_path = os.path.join(self.private_path, 'commit-cache')
            self._commit_cache_v = CommitLogCache(commit_cache_path)
            return self._commit_cache_v

    def _command(self, cmd, input=None, **kwargs):
        kwargs.setdefault('cwd', self.path)
        try:
            output = subprocess.check_output(cmd, **kwargs)
            return output
        except AttributeError:  # subprocess.check_output added in python 2.7
            kwargs.setdefault('stdout', subprocess.PIPE)
            p = subprocess.Popen(cmd, **kwargs)
            stdout, stderr = p.communicate()
            if p.returncode != 0:
                raise subprocess.CalledProcessError(p.returncode, cmd)
            return stdout

    @classmethod
    def cleanPath(cls, path):
        path = path.lstrip('/')
        path = multislash_rx.sub('/', path)
        return path

    @abstractmethod
    def canonical_rev(self, rev):
        """Get the canonical revision identifier"""
        raise NotImplementedError

    @abstractmethod
    def ls(
        self, rev, path, recursive=False, recursive_dirs=False,
        directory=False, report=()
    ):
        """List directory or file

        :param rev: The revision to use.
        :param path: The path to list. May start with a '/' or not. Directories
                                 may end with a '/' or not.
        :param recursive: Recursively list files in subdirectories.
        :param recursive_dirs: Used when recursive=True, also list directories.
        :param directory: If path is a directory, list path itself instead of its
                                            contents.
        :param report: A list or tuple of extra attributes to return that may
                                     require extra processing. Recognized values are 'size',
                                     'target', 'executable', and 'commit'.

        Returns a list of dictionaries with the following keys:

        **type**
            The type of the file: 'f' for file, 'd' for directory, 'l' for
            symlink.
        **name**
            The name of the file. Not present if directory=True.
        **size**
            The size of the file. Only present for files when 'size' is in
            report.
        **target**
            The target of the symlink. Only present for symlinks when
            'target' is in report.
        **executable**
            True if the file is executable, False otherwise.    Only present
            for files when 'executable' is in report.

        Raises PathDoesNotExist if the path does not exist.

        """
        raise NotImplementedError

    @abstractmethod
    def cat(self, rev, path):
        """Get file contents

        :param rev: The revision to use.
        :param path: The path to the file. Must be a file.

        Returns the file contents as a string.

        Raises PathDoesNotExist if the path does not exist.
        Raises BadFileType if the path is not a file.

        """
        raise NotImplementedError

    @abstractmethod
    def readlink(self, rev, path):
        """Get symbolic link target

        :param rev: The revision to use.
        :param path: The path to the file. Must be a symbolic link.

        Returns the target of the symbolic link as a string.

        Raises PathDoesNotExist if the path does not exist.
        Raises BadFileType if the path is not a symbolic link.

        """
        raise NotImplementedError

    @abstractmethod
    def branches(self):
        """Get list of branches
        """
        raise NotImplementedError

    @abstractmethod
    def tags(self):
        """Get list of tags
        """
        raise NotImplementedError

    @abstractmethod
    def heads(self):
        """Get list of heads
        """
        raise NotImplementedError

    @abstractmethod
    def empty(self):
        """Test if the repository contains any commits

        Commits that exist by default (e.g. a zero commit) are not counted.
        """
        return NotImplementedError

    @abstractmethod
    def __contains__(self, rev):
        """Test if the repository contains the specified revision
        """
        return NotImplementedError

    @abstractmethod
    def __len__(self):
        """Returns the number of commits in the repository

        Commits that exist by default (e.g. a zero commit) are not counted.
        """
        return NotImplementedError

    @abstractmethod
    def log(
        self, revrange=None, limit=None, firstparent=False, merges=None,
        path=None, follow=False
    ):
        """Get commit logs

        :param revrange: Either a single revision or a range of revisions as a 2
                                         element list or tuple.
        :param limit: Limit the number of log entries.
        :param firstparent: Only follow the first parent of merges.
        :param merges: True means only merges, False means no merges, None means
                                     both merges and non-merges.
        :param path:     Only match commits containing changes on this path.
        :param follow: Follow file history across renames.

        If revrange is None, return a list of all log entries in reverse
        chronological order.

        If revrange is a single revision, return a single log entry.

        If revrange is a 2 element list [A,B] or tuple (A,B), return a list of log
        entries starting at B and following that branch back to A or one of its
        ancestors (not inclusive. If A is None, follow branch B back to the
        beginning of history. If B is None, list all descendants in reverse
        chronological order.

        """
        raise NotImplementedError

    @abstractmethod
    def changed(self, rev):
        """Files that changed from the rev's parent(s)

        Returns a list of FileChangeInfo items.

        """
        raise NotImplementedError

    @abstractmethod
    def pdiff(self, rev):
        """Diff from the rev's parent(s)

        Returns a string containing the unified diff that the rev introduces with
        a prefix of one (suitable for input to patch -p1).

        """
        raise NotImplementedError

    @abstractmethod
    def diff(self, rev_a, rev_b, path=None):
        """Diff of two revisions

        Returns a string containing the unified diff from rev_a to rev_b with a
        prefix of one (suitable for input to patch -p1). If path is not None, only
        return the diff for that file.

        """
        raise NotImplementedError

    @abstractmethod
    def ancestor(self, rev1, rev2):
        """Find most recent common ancestor of two revisions
        """
        raise NotImplementedError

    @abstractmethod
    def blame(self, rev, path):
        """Blame (a.k.a. annotate, praise) a file

        Returns a list of BlameInfo objects in file order.

        Raises PathDoesNotExist if the path does not exist.
        Raises BadFileType if the path is not a file.

        """
        raise NotImplementedError



# Copyright (c) 2013, Clemson University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the {organization} nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

__version__ = '1.3.1'


def create(path, vcs):
    """Create a new repository

    vcs is either 'git', 'hg', or 'svn'

    """
    from .common import UnknownVCSType
    if vcs == 'git':
        from .git import GitRepo
        cls = GitRepo
    elif vcs == 'hg':
        from .hg import HgRepo
        cls = HgRepo
    elif vcs == 'svn':
        from .svn import SvnRepo
        cls = SvnRepo
    else:
        raise UnknownVCSType(vcs)
    return cls.create(path)


def open(path, vcs=None):
    """Open an existing repository

    vcs can be specified to avoid auto-detection of repository type

    """
    import os
    from .common import UnknownVCSType
    assert os.path.isdir(path), path + ' is not a directory'
    if vcs == 'git':
        from .git import GitRepo
        cls = GitRepo
    elif vcs == 'hg':
        from .hg import HgRepo
        cls = HgRepo
    elif vcs == 'svn':
        from .svn import SvnRepo
        cls = SvnRepo
    elif os.path.isdir(os.path.join(path, '.git')):
        from .git import GitRepo
        cls = GitRepo
    elif os.path.isdir(os.path.join(path, '.hg')):
        from .hg import HgRepo
        cls = HgRepo
    elif (
        os.path.isfile(os.path.join(path, 'config')) and
        os.path.isdir(os.path.join(path, 'objects')) and
        os.path.isdir(os.path.join(path, 'refs')) and
        os.path.isdir(os.path.join(path, 'branches'))
    ):
        from .git import GitRepo
        cls = GitRepo
    elif (
        os.path.isfile(os.path.join(path, 'format')) and
        os.path.isdir(os.path.join(path, 'conf')) and
        os.path.isdir(os.path.join(path, 'db')) and
        os.path.isdir(os.path.join(path, 'locks'))
    ):
        from .svn import SvnRepo
        cls = SvnRepo
    else:
        raise UnknownVCSType(path)
    return cls(path)



# Copyright (c) 2013, Clemson University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the {organization} nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import collections
import errno
import fcntl
import os


class HashDict(collections.MutableMapping):
    """A dictionary-like object for hex keys and string values that is stored
    on-disk and is multi-process safe.
    """

    def __init__(self, path, mode=0o666):
        self.path = path
        self.mode = mode
        self.dirmode = mode | (mode >> 1) & 0o111 | (mode >> 2) & 0o111
        try:
            os.mkdir(path, self.dirmode)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

    def __contains__(self, key):
        int(key, 16)
        p = os.path.join(self.path, key[:2], key[2:])
        return os.path.isfile(p)

    def __getitem__(self, key):
        int(key, 16)
        p = os.path.join(self.path, key[:2], key[2:])
        try:
            with open(p, 'r') as f:
                fcntl.lockf(f, fcntl.LOCK_SH)
                return f.read()
        except IOError as e:
            if e.errno == errno.ENOENT:
                raise KeyError(key)
            raise

    def __setitem__(self, key, value):
        int(key, 16)
        d = os.path.join(self.path, key[:2])
        p = os.path.join(d, key[2:])
        try:
            os.mkdir(d, self.dirmode)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
        fd = None
        try:
            fd = os.open(p, os.O_WRONLY | os.O_CREAT, self.mode)
            fcntl.lockf(fd, fcntl.LOCK_EX)
            os.ftruncate(fd, 0)
            with os.fdopen(fd, 'w') as f:
                fd = None
                f.write(value)
        finally:
            if fd is not None:
                os.close(fd)

    def __delitem__(self, key):
        int(key, 16)
        d = os.path.join(self.path, key[:2])
        p = os.path.join(d, key[2:])
        try:
            os.unlink(p)
        except OSError as e:
            if e.errno == errno.ENOENT:
                raise KeyError(key)
            raise

    def __iter__(self):
        for d in os.listdir(self.path):
            try:
                int(d, 16)
            except ValueError:
                continue
            p = os.path.join(self.path, d)
            for k in os.listdir(p):
                try:
                    int(k, 16)
                except ValueError:
                    continue
                if os.path.isfile(os.path.join(p, k)):
                    yield d + k

    def __len__(self):
        return len(list(self.__iter__()))



# Copyright (c) 2013, Clemson University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the {organization} nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import collections
import fnmatch
import re
import subprocess
import sys
from .common import *

DIFF = 'diff'
SVN = 'svn'
SVNADMIN = 'svnadmin'
SVNLOOK = 'svnlook'

head_rev_rx = re.compile(r'^(?=.)(?P<head>\D[^:]*)?:?(?P<rev>\d+)?$')
mergeinfo_rx = re.compile(r'^(?P<head>.+):(?P<minrev>\d+)(?:-(?P<maxrev>\d+))$')
changed_copy_info_rx = re.compile(r'^[ ]{4}\(from (?P<src>.+)\)$')

HistoryEntry = collections.namedtuple('HistoryEntry', 'rev path')


class SvnRepo(VCSRepo):
    """A Subversion repository

    Unless otherwise specified, valid revisions are:

    - an integer (ex: 194)
    - an integer as a string (ex: "194")
    - a branch or tag name (ex: "HEAD", "trunk", "branches/branch1")
    - a branch or tag name at a specific revision (ex: "trunk:194")

    Revisions have the following meanings:

    - HEAD always maps to the root of the repository (/)
    - Anything else (ex: "trunk", "branches/branch1") maps to the corresponding
        path in the repository
    - The youngest revision is assumed unless a revision is specified

    For example, the following code will list the contents of the directory
    branches/branch1/src from revision 194:

            >>> repo = SvnRepo(path)
            >>> repo.ls('branches/branch1:194', 'src')

    Branches and tags are detected in branches() and tags() by looking at the
    paths specified in repo.branch_glob and repo.tag_glob.    The default values
    for these variables will detect the following repository layout:

    - /trunk - the main development branch
    - /branches/* - branches
    - /tags/* - tags

    If a repository does not fit this layout, everything other than branch and
    tag detection will work as expected.

    """

    @classmethod
    def create(cls, path):
        """Create a new repository"""
        cmd = [SVNADMIN, 'create', path]
        subprocess.check_call(cmd)
        return cls(path)

    @classmethod
    def cleanPath(cls, path):
        path = multislash_rx.sub('/', path)
        if not path.startswith('/'):
            path = '/' + path
        return path

    def __init__(self, path):
        super(SvnRepo, self).__init__(path)
        self.branch_glob = ['/trunk/', '/branches/*/']
        self.tag_glob = ['/tags/*/']

    @property
    def private_path(self):
        """Get the path to a directory which can be used to store arbitrary data

        This directory should not conflict with any of the repository internals.
        The directory should be created if it does not already exist.

        """
        import os
        path = os.path.join(self.path, '.private')
        try:
            os.mkdir(path)
        except OSError as e:
            import errno
            if e.errno != errno.EEXIST:
                raise
        return path

    def _proplist(self, rev, path):
        cmd = [SVNLOOK, 'proplist', '-r', rev, '.', path or '--revprop']
        output = self._command(cmd).decode()
        return [x.strip() for x in output.splitlines()]

    def proplist(self, rev, path=None):
        """List Subversion properties of the path"""
        rev, prefix = self._maprev(rev)
        if path is None:
            return self._proplist(str(rev), None)
        else:
            path = type(self).cleanPath(prefix + path)
            return self._proplist(str(rev), path)

    def _propget(self, prop, rev, path):
        cmd = [SVNLOOK, 'propget', '-r', rev, '.', prop, path or '--revprop']
        return self._command(cmd).decode()

    def propget(self, prop, rev, path=None):
        """Get Subversion property value of the path"""
        rev, prefix = self._maprev(rev)
        if path is None:
            return self._propget(prop, str(rev), None)
        else:
            path = type(self).cleanPath(prefix + path)
            return self._propget(prop, str(rev), path)

    def _mergeinfo(self, rev, path):
        revstr = str(rev)
        if 'svn:mergeinfo' not in self._proplist(revstr, path):
            return []
        results = []
        mergeinfo = self._propget('svn:mergeinfo', revstr, path)
        for line in mergeinfo.splitlines():
            m = mergeinfo_rx.match(line)
            assert m
            head, minrev, maxrev = m.group('head', 'minrev', 'maxrev')
            minrev = int(minrev)
            maxrev = int(maxrev or minrev)
            results.append((head, minrev, maxrev))
        return results

    def _maprev(self, rev):
        if isinstance(rev, int):
            return (rev, '/')
        m = head_rev_rx.match(rev)
        assert m, 'invalid rev'
        head, rev = m.group('head', 'rev')
        if rev:
            rev = int(rev)
        else:
            rev = self.youngest()
        if head is None:
            return (rev, '/')
        elif head == 'HEAD':
            return (rev, '/')
        else:
            return (rev, '/' + head)

    def canonical_rev(self, rev):
        try:
            types = (str, unicode)
        except NameError:
            types = str
        if isinstance(rev, int):
            return rev
        elif isinstance(rev, types) and rev.isdigit():
            return int(rev)
        else:
            rev, prefix = self._maprev(rev)
            return rev

    def ls(
        self, rev, path, recursive=False, recursive_dirs=False,
        directory=False, report=()
    ):
        rev, prefix = self._maprev(rev)
        revstr = str(rev)
        path = type(self).cleanPath(prefix + path)
        forcedir = False
        if path.endswith('/'):
            forcedir = True
            if path != '/':
                path = path.rstrip('/')
        if path == '/':
            if directory:
                entry = attrdict(path='/', type='d')
                if 'commit' in report:
                    entry.commit = self._history(revstr, '/', 1)[0].rev
                return [entry]
            ltrim = 1
            prefix = '/'
        else:
            ltrim = len(path) + 1
            prefix = path + '/'

        cmd = [SVNLOOK, 'tree', '-r', revstr, '--full-paths']
        if not recursive:
            cmd.append('--non-recursive')
        cmd.extend(['.', path])
        p = subprocess.Popen(
            cmd, cwd=self.path, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        output, stderr = p.communicate()
        if p.returncode != 0:
            stderr = stderr.decode()
            if p.returncode == 1 and 'File not found' in stderr:
                raise PathDoesNotExist(rev, path)
            raise subprocess.CalledProcessError(p.returncode, cmd, stderr)

        results = []
        lines = output.decode(self.encoding, 'replace').splitlines()
        if forcedir and not lines[0].endswith('/'):
            raise PathDoesNotExist(rev, path)
        if lines[0].endswith('/'):
            if directory:
                lines = lines[:1]
            else:
                lines = lines[1:]
        for name in lines:
            entry_name = name[ltrim:]
            entry = attrdict(path=name.strip('/'))
            if name.endswith('/'):
                if recursive and not recursive_dirs:
                    continue
                entry.type = 'd'
                entry_name = entry_name.rstrip('/')
            else:
                proplist = self._proplist(revstr, name)
                if 'svn:special' in proplist:
                    link = self._cat(revstr, name).decode(self.encoding, 'replace')
                    link = link.split(None, 1)
                    if len(link) == 2 and link[0] == 'link':
                        entry.type = 'l'
                        if 'target' in report:
                            entry.target = link[1]
                if 'type' not in entry:
                    entry.type = 'f'
                    if 'executable' in report:
                        entry.executable = 'svn:executable' in proplist
                    if 'size' in report:
                        entry.size = len(self._cat(revstr, name))
            if entry_name:
                entry.name = entry_name
            if 'commit' in report:
                entry.commit = self._history(revstr, name, 1)[0].rev
            results.append(entry)
        return results

    def _cat(self, rev, path):
        cmd = [SVNLOOK, 'cat', '-r', rev, '.', path.encode(self.encoding)]
        return self._command(cmd)

    def cat(self, rev, path):
        rev, prefix = self._maprev(rev)
        path = type(self).cleanPath(prefix + path)
        ls = self.ls(rev, path, directory=True)
        assert len(ls) == 1
        if ls[0].get('type') != 'f':
            raise BadFileType(rev, path)
        return self._cat(str(rev), path)

    def _readlink(self, rev, path):
        output = self._cat(rev, path)
        link = output.decode(self.encoding, 'replace').split(None, 1)
        assert len(link) == 2 and link[0] == 'link'
        return link[1]

    def readlink(self, rev, path):
        rev, prefix = self._maprev(rev)
        path = type(self).cleanPath(prefix + path)
        ls = self.ls(rev, path, directory=True)
        assert len(ls) == 1
        if ls[0].get('type') != 'l':
            raise BadFileType(rev, path)
        return self._readlink(str(rev), path)

    def youngest(self):
        cmd = [SVNLOOK, 'youngest', '.']
        return int(self._command(cmd))

    def _heads(self, globs):
        root = {}
        for glob in globs:
            n = root
            for p in glob.strip('/').split('/'):
                n = n.setdefault(p, {})
        youngest = self.youngest()
        results = []

        def match(n, path):
            for d in self.ls(youngest, path):
                if d.get('type') == 'd':
                    for k, v in n.items():
                        if fnmatch.fnmatchcase(d.name, k):
                            if path:
                                p = path + '/' + d.name
                            else:
                                p = d.name
                            if v:
                                match(v, p)
                            else:
                                results.append(p)
        match(root, '')
        return results

    def branches(self):
        return ['HEAD'] + self._heads(self.branch_glob)

    def tags(self):
        return self._heads(self.tag_glob)

    def heads(self):
        return ['HEAD'] + self._heads(self.branch_glob + self.tag_glob)

    def empty(self):
        cmd = [SVNLOOK, 'history', '.', '-l2']
        output = self._command(cmd)
        return len(output.splitlines()) < 4

    def __contains__(self, rev):
        rev, prefix = self._maprev(rev)
        cmd = [SVNLOOK, 'history', '.', prefix, '-l1', '-r', str(rev)]
        p = subprocess.Popen(
            cmd, cwd=self.path, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        stdout, stderr = p.communicate()
        return p.returncode == 0

    def __len__(self):
        cmd = [SVNLOOK, 'history', '.']
        output = self._command(cmd)
        return len(output.splitlines()) - 3

    def log(
        self, revrange=None, limit=None, firstparent=False, merges=None,
        path=None, follow=False
    ):
        if not (revrange is None or isinstance(revrange, (tuple, list))):
            # a single revision was given
            rev, prefix = self._maprev(revrange)
            h = self._history(rev, prefix, 1)
            rev = h[0].rev
            return self._logentry(rev, prefix)

        if revrange is None:
            results = self._history(self.youngest(), path or '/', limit)
        else:
            if revrange[1] is None:
                include = set()
                rev1 = self.youngest()
                for head in self.heads():
                    if head == 'HEAD':
                        continue
                    if path:
                        p = head + '/' + path.lstrip('/')
                    else:
                        p = type(self).cleanPath(head)
                    include.update(self._mergehistory(rev1, p, limit))
            else:
                rev1, prefix1 = self._maprev(revrange[1])
                if path:
                    p = type(self).cleanPath(prefix1 + '/' + path)
                else:
                    p = prefix1
                if firstparent:
                    include = self._history(rev1, p)
                else:
                    include = self._mergehistory(rev1, p, limit)

            if revrange[0] is None:
                results = include
            else:
                rev0, prefix0 = self._maprev(revrange[0])
                exclude = self._mergehistory(rev0, prefix0)
                results = include - exclude

            results = sorted(results, key=lambda x: x.rev, reverse=True)

        results = map(lambda x: self._logentry(x.rev, x.path), results)
        if merges is not None:
            if merges:
                results = filter(lambda x: len(x.parents) > 1, results)
            else:
                results = filter(lambda x: len(x.parents) <= 1, results)
        return list(results)

    def _logentry(self, rev, path, history=None):
        import hashlib
        revstr = str(rev)
        cmd = [SVNLOOK, 'info', '.', '-r', revstr]
        cachekey = hashlib.sha1(revstr.encode()).hexdigest()
        entry = self._commit_cache.get(cachekey)
        if entry:
            entry._cached = True
            return entry
        output = self._command(cmd).decode(self.encoding, 'replace')
        author, date, logsize, message = output.split('\n', 3)
        date = parse_isodate(date)
        if history is None:
            history = self._history(rev, path, 2)
        parents = []
        if len(history) > 1:
            prev = history[1].rev
            if path == '/':
                parents.append(prev)
            else:
                parents.append('%s:%d' % (path, prev))
            for head, minrev, maxrev in self._mergeinfo(rev, path):
                if prev < maxrev:
                    h = self._history(maxrev, head, 1)
                    if head == '/':
                        parents.append(h[0].rev)
                    else:
                        parents.append('%s:%d' % (head, h[0].rev))
        entry = CommitLogEntry(rev, parents, date, author, message)
        if cachekey not in self._commit_cache:
            self._commit_cache[cachekey] = entry
        return entry

    def pdiff(self, rev):
        rev, prefix = self._maprev(rev)
        if rev == 0:
            return ''
        cmd = [SVNLOOK, 'diff', '.', '-r', str(rev)]
        output = self._command(cmd)
        _output = b''
        for line in output.splitlines(True):
            if line.startswith(b'--- '):
                line = b'--- a/' + line[4:]
            if line.startswith(b'+++ '):
                line = b'+++ b/' + line[4:]
            _output += line
        output = _output
        return output

    def diff(self, rev_a, rev_b, path=None):
        import os
        import shutil
        import tempfile

        rev_a, prefix_a = self._maprev(rev_a)
        rev_b, prefix_b = self._maprev(rev_b)
        tmpdir = tempfile.mkdtemp(prefix='anyvcs-svn-diff.')
        try:
            path_a = os.path.join(tmpdir, 'a')
            path_b = os.path.join(tmpdir, 'b')
            url_a = 'file://%s/%s@%d' % (self.path, prefix_a, rev_a)
            url_b = 'file://%s/%s@%d' % (self.path, prefix_b, rev_b)
            cmd = [SVN, 'export', '-q', url_a, path_a]
            subprocess.check_call(cmd)
            cmd = [SVN, 'export', '-q', url_b, path_b]
            subprocess.check_call(cmd)
            if path is None:
                cmd = [DIFF, '-urN', 'a', 'b']
            else:
                path = type(self).cleanPath(path)
                cmd = [DIFF, '-urN', 'a' + path, 'b' + path]
            p = subprocess.Popen(cmd, cwd=tmpdir, stdout=subprocess.PIPE)
            stdout, stderr = p.communicate()
            if p.returncode not in (0, 1):
                raise subprocess.CalledProcessError(p.returncode, cmd, stdout)
            return stdout
        finally:
            shutil.rmtree(tmpdir)

    def changed(self, rev):
        rev, prefix = self._maprev(rev)
        if rev == 0:
            return []
        cmd = [SVNLOOK, 'changed', '.', '-r', str(rev), '--copy-info']
        output = self._command(cmd).decode(self.encoding, 'replace')
        lines = output.splitlines()
        lines.reverse()
        results = []
        while lines:
            line = lines.pop()
            status = line[:3]
            path = line[4:].lstrip('/')
            copy = None
            if status.endswith('+'):
                line = lines.pop()
                m = changed_copy_info_rx.match(line)
                assert m
                copy = m.group('src')
            entry = FileChangeInfo(path, str(status), copy)
            results.append(entry)
        return results

    def _history(self, rev, path, limit=None):
        cmd = [SVNLOOK, 'history', '.', '-r', str(rev), path]
        if limit is not None:
            cmd.extend(['-l', str(limit)])
        output = self._command(cmd).decode(self.encoding, 'replace')
        results = []
        for line in output.splitlines()[2:]:
            r, p = line.split(None, 1)
            results.append(HistoryEntry(int(r), p))
        return results

    def _mergehistory(self, rev, path, limit=None):
        results = set(self._history(rev, path, limit))
        for head, minrev, maxrev in self._mergeinfo(rev, path):
            l = maxrev - minrev + 1
            if limit is not None:
                l = min(l, limit)
            h = self._history(maxrev, head, l)
            for r, p in h:
                if r < minrev:
                    break
                results.add(HistoryEntry(r, p))
        return results

    def ancestor(self, rev1, rev2):
        rev1, prefix1 = self._maprev(rev1)
        rev2, prefix2 = self._maprev(rev2)
        prefix1 = type(self).cleanPath(prefix1)
        if prefix1 != '/':
            prefix1 = prefix1.rstrip('/')
        prefix2 = type(self).cleanPath(prefix2)
        if prefix2 != '/':
            prefix2 = prefix2.rstrip('/')

        self.ls(rev1, prefix1, directory=True)
        self.ls(rev2, prefix2, directory=True)

        minrev = min(rev1, rev2)
        if prefix1 == prefix2:
            return '%s:%d' % (prefix1, minrev)

        history1 = self._history(minrev, prefix1)
        history2 = self._history(minrev, prefix2)

        youngest = HistoryEntry(0, '/')

        for head, minrev, maxrev in self._mergeinfo(rev1, prefix1):
            for h in history2:
                if h.rev < minrev or h.rev < youngest.rev:
                    break
                if h.path == head and minrev <= h.rev <= maxrev:
                    youngest = h

        for head, minrev, maxrev in self._mergeinfo(rev2, prefix2):
            for h in history1:
                if h.rev < minrev or h.rev < youngest.rev:
                    break
                if h.path == head and minrev <= h.rev <= maxrev:
                    youngest = h

        if youngest.rev > 0:
            return '%s:%d' % (youngest.path, youngest.rev)

        i1 = 0
        i2 = 0
        len1 = len(history1)
        len2 = len(history2)
        while i1 < len1 and i2 < len2:
            if history1[i1].rev < history2[i2].rev:
                i2 += 1
            elif history1[i1].rev > history2[i2].rev:
                i1 += 1
            else:
                if history1[i1].path == history2[i2].path:
                    return '%s:%d' % (history1[i1].path, history1[i1].rev)
                else:
                    i1 += 1
                    i2 += 1

        return None

    def _blame(self, rev, path):
        import os
        import xml.etree.ElementTree as ET
        url = 'file://' + os.path.abspath(self.path) + path
        cmd = [SVN, 'blame', '--xml', '-r', rev, url]
        output = self._command(cmd)
        tree = ET.fromstring(output)
        results = []
        cat = self._cat(rev, path)
        target = tree.find('target')
        try:
            iter = target.iter('entry')
        except AttributeError:  # added in python 2.7
            iter = target.getiterator('entry')
        for entry, text in zip(iter, cat.splitlines()):
            commit = entry.find('commit')
            rev = int(commit.attrib.get('revision'))
            author = commit.find('author').text
            date = commit.find('date').text
            date = parse_isodate(date)
            results.append(BlameInfo(rev, author, date, text))
        return results

    def blame(self, rev, path):
        rev, prefix = self._maprev(rev)
        path = type(self).cleanPath(prefix + path)
        ls = self.ls(rev, path, directory=True)
        assert len(ls) == 1
        if ls[0].get('type') != 'f':
            raise BadFileType(rev, path)
        return self._blame(str(rev), path)

    def dump(
        self, stream, progress=None, lower=None, upper=None,
        incremental=False, deltas=False
    ):
        """Dump the repository to a dumpfile stream.

        :param stream: A file stream to which the dumpfile is written
        :param progress: A file stream to which progress is written
        :param lower: Must be a numeric version number
        :param upper: Must be a numeric version number

        See ``svnadmin help dump`` for details on the other arguments.

        """
        cmd = [SVNADMIN, 'dump', '.']
        if progress is None:
            cmd.append('-q')
        if lower is not None:
            cmd.append('-r')
            if upper is None:
                cmd.append(str(int(lower)))
            else:
                cmd.append('%d:%d' % (int(lower), int(upper)))
        if incremental:
            cmd.append('--incremental')
        if deltas:
            cmd.append('--deltas')
        p = subprocess.Popen(cmd, cwd=self.path, stdout=stream, stderr=progress)
        p.wait()
        if p.returncode != 0:
            raise subprocess.CalledProcessError(p.returncode, cmd)

    def load(
        self, stream, progress=None, ignore_uuid=False, force_uuid=False,
        use_pre_commit_hook=False, use_post_commit_hook=False, parent_dir=None
    ):
        """Load a dumpfile stream into the repository.

        :param stream: A file stream from which the dumpfile is read
        :param progress: A file stream to which progress is written

        See ``svnadmin help load`` for details on the other arguments.

        """
        cmd = [SVNADMIN, 'load', '.']
        if progress is None:
            cmd.append('-q')
        if ignore_uuid:
            cmd.append('--ignore-uuid')
        if force_uuid:
            cmd.append('--force-uuid')
        if use_pre_commit_hook:
            cmd.append('--use-pre-commit-hook')
        if use_post_commit_hook:
            cmd.append('--use-post-commit-hook')
        if parent_dir:
            cmd.extend(['--parent-dir', parent_dir])
        p = subprocess.Popen(
            cmd, cwd=self.path, stdin=stream, stdout=progress,
            stderr=subprocess.PIPE
        )
        stderr = p.stderr.read()
        p.wait()
        if p.returncode != 0:
            raise subprocess.CalledProcessError(p.returncode, cmd, stderr)



# Copyright (c) 2013, Clemson University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the {organization} nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import datetime
import os
import re
import subprocess
from .common import *

HG = 'hg'

canonical_rev_rx = re.compile(r'^[0-9a-f]{40}$')
manifest_rx = re.compile(r'^(?P<object>[0-9a-f]{40}) (?P<mode>[0-7]{3}) (?P<type>.) (?P<name>.+)$')
parse_heads_rx = re.compile(r'^(?P<name>.+?)\s+(?P<rev>-?\d+):(?P<nodeid>[0-9a-f]+)', re.I)
bookmarks_rx = re.compile(r'^\s+(?:\*\s+)?(?P<name>.+?)\s+(?P<rev>\d+):(?P<nodeid>[0-9a-f]+)', re.I)
annotate_rx = re.compile(r'^(?P<author>.*)\s+(?P<rev>\d+):\s')


def parent_dirs(path):
    ds = path.find('/')
    while ds != -1:
        yield path[:ds]
        ds = path.find('/', ds + 1)


def parse_hgdate(datestr):
    ts, tzoffset = datestr.split(None, 1)
    date = datetime.datetime.fromtimestamp(float(ts))
    return date.replace(tzinfo=UTCOffset(-int(tzoffset) / 60))


class HgRepo(VCSRepo):
    """A Mercurial repository

    Valid revisions are anything that Mercurial considers as a revision.

    """

    @classmethod
    def create(cls, path):
        """Create a new repository"""
        cmd = [HG, 'init', path]
        subprocess.check_call(cmd)
        return cls(path)

    @property
    def private_path(self):
        """Get the path to a directory which can be used to store arbitrary data

        This directory should not conflict with any of the repository internals.
        The directory should be created if it does not already exist.

        """
        path = os.path.join(self.path, '.hg', '.private')
        try:
            os.mkdir(path)
        except OSError as e:
            import errno
            if e.errno != errno.EEXIST:
                raise
        return path

    @property
    def _object_cache(self):
        try:
            return self._object_cache_v
        except AttributeError:
            object_cache_path = os.path.join(self.private_path, 'object-cache')
            self._object_cache_v = HashDict(object_cache_path)
            return self._object_cache_v

    def canonical_rev(self, rev):
        if isinstance(rev, str) and canonical_rev_rx.match(rev):
            return rev
        else:
            cmd = [HG, 'log', '--template={node}', '-r', str(rev)]
            return self._command(cmd).decode()

    def _revnum(self, rev):
        if isinstance(rev, int):
            return rev
        elif isinstance(rev, str) and rev.isdigit():
            return int(rev)
        else:
            cmd = [HG, 'log', '--template={rev}', '-r', str(rev)]
            return int(self._command(cmd))

    def _ls(
        self, rev, path, recursive=False, recursive_dirs=False, directory=False
    ):
        forcedir = False
        if path.endswith('/'):
            forcedir = True
            path = path.rstrip('/')
        if path == '':
            ltrim = 0
            prefix = ''
        else:
            ltrim = len(path) + 1
            prefix = path + '/'
        cmd = [HG, 'manifest', '--debug', '-r', rev]
        output = self._command(cmd).decode(self.encoding, 'replace')
        if not output:
            return

        dirs = set()
        exists = False
        for line in output.splitlines():
            m = manifest_rx.match(line)
            assert m, 'unexpected output: ' + line
            t, name, objid = m.group('type', 'name', 'object')
            if name.startswith(prefix) or (not forcedir and name == path):
                if directory and name.startswith(prefix):
                    yield ('d', path, '', None)
                    return
                exists = True
                entry_name = name[ltrim:]
                if '/' in entry_name:
                    p = parent_dirs(entry_name)
                    if not recursive:
                        d = next(p)
                        if d not in dirs:
                            dirs.add(d)
                            yield ('d', prefix + d, d, None)
                        continue
                    if recursive_dirs:
                        for d in p:
                            if d not in dirs:
                                dirs.add(d)
                                yield ('d', prefix + d, d, None)
                yield (t, name, entry_name, objid)
        if not exists:
            raise PathDoesNotExist(rev, path)

    def ls(
        self, rev, path, recursive=False, recursive_dirs=False,
        directory=False, report=()
    ):
        revstr = str(rev)
        path = type(self).cleanPath(path)
        if path == '':
            if directory:
                entry = attrdict(path='/', type='d')
                if 'commit' in report:
                    entry.commit = self.canonical_rev(revstr)
                return [entry]

        if 'commit' in report:
            import fcntl
            import tempfile
            files_cache_path = os.path.join(self.private_path, 'files-cache.log')
            with open(files_cache_path, 'a+') as files_cache:
                fcntl.lockf(files_cache, fcntl.LOCK_EX, 0, 0, os.SEEK_CUR)
                files_cache.seek(0)
                log = files_cache.read().split('\0')
                assert log.pop() == ''
                if log:
                    startlog = int(log[-1].splitlines()[0]) + 1
                    if startlog >= len(self):
                        startlog = None
                else:
                    startlog = 0
                if startlog is not None:
                    with tempfile.NamedTemporaryFile() as style:
                        style.write((
                            r"changeset = '{rev}\n{node}\n{parents}\n{files}\0'" '\n'
                            r"parent = '{rev} '" '\n'
                            r"file = '{file|escape}\n'" '\n'
                        ).encode())
                        style.flush()
                        cmd = [HG, 'log', '--style', style.name, '-r', '%d:' % startlog]
                        output = self._command(cmd).decode(self.encoding, 'replace')
                        files_cache.write(output)
                        extend = output.split('\0')
                        assert extend.pop() == ''
                        log.extend(extend)

        results = []
        lookup_commit = {}
        for t, fullpath, name, objid in self._ls(revstr, path, recursive, recursive_dirs, directory):
            entry = attrdict(path=fullpath)
            if name:
                entry.name = name
            if t == 'd':
                entry.type = 'd'
            elif t in ' *':
                entry.type = 'f'
                if 'executable' in report:
                    entry.executable = t == '*'
                if 'size' in report:
                    entry.size = len(self._cat(revstr, name))
            elif t == '@':
                entry.type = 'l'
                if 'target' in report:
                    entry.target = self._cat(revstr, name).decode(self.encoding, 'replace')
            else:
                assert False, 'unexpected output: ' + line
            if 'commit' in report:
                lookup = True
                if objid:
                    try:
                        entry.commit = self._object_cache[objid]
                        entry._commit_cached = True
                        lookup = False
                    except KeyError:
                        pass
                if lookup:
                    p = type(self).cleanPath(path + '/' + name)
                    lookup_commit[p] = (entry, objid)
            results.append(entry)

        if 'commit' in report:
            import heapq
            ancestors = [-self._revnum(revstr)]
            while ancestors and lookup_commit:
                r = -heapq.heappop(ancestors)
                lines = log[r].splitlines()
                parents = lines[2]
                if parents:
                    for x in parents.split():
                        x = int(x)
                        if x != -1:
                            if -x not in ancestors:
                                heapq.heappush(ancestors, -x)
                elif r > 0:
                    x = r - 1
                    if x not in ancestors:
                        heapq.heappush(ancestors, -x)
                for p in list(lookup_commit):
                    prefix = p.rstrip('/') + '/'
                    for l in lines[3:]:
                        if l == p or l.startswith(prefix):
                            commit = str(lines[1])
                            entry, objid = lookup_commit[p]
                            entry.commit = commit
                            if objid:
                                self._object_cache[objid] = commit
                            del lookup_commit[p]
                            break

        return results

    def _cat(self, rev, path):
        cmd = [HG, 'cat', '-r', rev, path.encode(self.encoding)]
        return self._command(cmd)

    def cat(self, rev, path):
        path = type(self).cleanPath(path)
        ls = self.ls(rev, path, directory=True)
        assert len(ls) == 1
        if ls[0].get('type') != 'f':
            raise BadFileType(rev, path)
        return self._cat(str(rev), path)

    def readlink(self, rev, path):
        path = type(self).cleanPath(path)
        ls = self.ls(rev, path, directory=True)
        assert len(ls) == 1
        if ls[0].get('type') != 'l':
            raise BadFileType(rev, path)
        return self._cat(str(rev), path).decode(self.encoding, 'replace')

    def _parse_heads(self, cmd):
        output = self._command(cmd).decode(self.encoding, 'replace')
        results = []
        for line in output.splitlines():
            m = parse_heads_rx.match(line)
            assert m, 'unexpected output: ' + line
            results.append(m.group('name'))
        return results

    def branches(self):
        cmd = [HG, 'branches']
        return self._parse_heads(cmd)

    def tags(self):
        cmd = [HG, 'tags']
        return self._parse_heads(cmd)

    def bookmarks(self):
        """Get list of bookmarks"""
        cmd = [HG, 'bookmarks']
        output = self._command(cmd).decode(self.encoding, 'replace')
        if output.startswith('no bookmarks set'):
            return []
        results = []
        for line in output.splitlines():
            m = bookmarks_rx.match(line)
            assert m, 'unexpected output: ' + line
            results.append(m.group('name'))
        return results

    def heads(self):
        return self.branches() + self.tags() + self.bookmarks()

    def empty(self):
        cmd = [HG, 'log', '--template=a', '-l1']
        output = self._command(cmd)
        return len(output) == 0

    def __contains__(self, rev):
        cmd = [HG, 'log', '--template=a', '-r', str(rev)]
        p = subprocess.Popen(
            cmd, cwd=self.path, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        stdout, stderr = p.communicate()
        return p.returncode == 0

    def __len__(self):
        cmd = [HG, 'id', '-n', '-r', 'tip']
        output = self._command(cmd)
        return int(output) + 1

    def log(
        self, revrange=None, limit=None, firstparent=False, merges=None,
        path=None, follow=False
    ):
        cmd = [
            HG, 'log', '--debug', '--template={node}\\0{parents}\\0'
            '{date|hgdate}\\0{author|nonempty}'
            '\\0{desc|tabindent|nonempty}\\0\\0']
        if limit is not None:
            cmd.append('-l' + str(limit))
        if firstparent:
            cmd.append('--follow-first')
        if merges is not None:
            if merges:
                cmd.append('--only-merges')
            else:
                cmd.append('--no-merges')
        single = False
        if revrange is None:
            pass
        elif isinstance(revrange, (tuple, list)):
            if revrange[0] is None:
                if revrange[1] is None:
                    pass
                else:
                    cmd.extend(['-r', 'reverse(ancestors(%s))' % revrange[1]])
            else:
                if revrange[1] is None:
                    cmd.extend(['-r', 'reverse(descendants(%s))' % revrange[0]])
                else:
                    cmd.extend(['-r', 'reverse(ancestors(%s))' % revrange[1], '--prune', str(revrange[0])])
        else:
            entry = self._commit_cache.get(self.canonical_rev(revrange))
            if entry:
                entry._cached = True
                return entry
            cmd.extend(['-r', str(revrange)])
            single = True
        if path:
            if follow:
                cmd.append('--follow')
            cmd.extend(['--', type(self).cleanPath(path)])
        output = self._command(cmd).decode(self.encoding, 'replace')

        results = []
        logs = output.split('\0\0')
        logs.pop()
        for log in logs:
            rev, parents, date, author, message = log.split('\0', 4)
            parents = [
                x[1] for x in filter(
                    lambda x: x[0] != '-1',
                    (x.split(':') for x in parents.split())
                )
            ]
            date = parse_hgdate(date)
            message = message.replace('\n\t', '\n')
            entry = CommitLogEntry(rev, parents, date, author, message)
            if rev not in self._commit_cache:
                self._commit_cache[rev] = entry
            if single:
                return entry
            results.append(entry)
        return results

    def changed(self, rev):
        cmd = [HG, 'status', '-C', '--change', str(rev)]
        output = self._command(cmd).decode(self.encoding, 'replace')
        results = []
        copy = None
        for line in reversed(output.splitlines()):
            if line.startswith(' '):
                copy = line.lstrip()
            else:
                status, path = line.split(None, 1)
                entry = FileChangeInfo(path, str(status), copy)
                results.append(entry)
                copy = None
        results.reverse()
        return results

    def pdiff(self, rev):
        cmd = [HG, 'log', '--template=a', '-p', '-r', str(rev)]
        return self._command(cmd)[1:]

    def diff(self, rev_a, rev_b, path=None):
        cmd = [HG, 'diff', '-r', rev_a, '-r', rev_b]
        if path is not None:
            cmd.extend(['--', type(self).cleanPath(path)])
        return self._command(cmd)

    def ancestor(self, rev1, rev2):
        cmd = [HG, 'log', '--template={node}', '-r', 'ancestor(%s, %s)' % (rev1, rev2)]
        output = self._command(cmd).decode()
        if output == '':
            return None
        else:
            return output

    def _blame(self, rev, path):
        cmd = [HG, 'annotate', '-unv', '-r', rev, '--', path]
        output = self._command(cmd).decode(self.encoding, 'replace')
        revs = {}
        results = []
        cat = self._cat(rev, path)
        for line, text in zip(output.splitlines(), cat.splitlines()):
            m = annotate_rx.match(line)
            assert m, 'unexpected output: ' + line
            rev, author = m.group('rev', 'author')
            try:
                rev, date = revs[rev]
            except KeyError:
                cmd = [HG, 'log', '--template={node}\n{date|hgdate}', '-r', rev]
                output = self._command(cmd).decode(self.encoding, 'replace')
                rev, date = output.split('\n', 1)
                date = parse_hgdate(date)
                revs[rev] = rev, date
            results.append(BlameInfo(rev, author, date, text))
        return results

    def blame(self, rev, path):
        path = type(self).cleanPath(path)
        ls = self.ls(rev, path, directory=True)
        assert len(ls) == 1
        if ls[0].get('type') != 'f':
            raise BadFileType(rev, path)
        return self._blame(str(rev), path)



# Copyright (c) 2013, Clemson University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the {organization} nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import re
import stat
import subprocess
from .common import *
from .hashdict import HashDict

GIT = 'git'

rev_rx = re.compile(r'^[0-9a-f]{40}$', re.IGNORECASE)
branch_rx = re.compile(r'^[*]?\s+(?P<name>.+)$')


class GitRepo(VCSRepo):
    """A git repository

    Valid revisions are anything that git considers as a revision.

    """

    @classmethod
    def create(cls, path, encoding='utf-8'):
        """Create a new bare repository"""
        cmd = [GIT, 'init', '--quiet', '--bare', path]
        subprocess.check_call(cmd)
        return cls(path, encoding)

    @property
    def private_path(self):
        """Get the path to a directory which can be used to store arbitrary data

        This directory should not conflict with any of the repository internals.
        The directory should be created if it does not already exist.

        """
        path = os.path.join(self.path, '.private')
        try:
            os.mkdir(path)
        except OSError as e:
            import errno
            if e.errno != errno.EEXIST:
                raise
        return path

    @property
    def _object_cache(self):
        try:
            return self._object_cache_v
        except AttributeError:
            object_cache_path = os.path.join(self.private_path, 'object-cache')
            self._object_cache_v = HashDict(object_cache_path)
            return self._object_cache_v

    def canonical_rev(self, rev):
        rev = str(rev)
        if rev_rx.match(rev):
            return rev
        else:
            cmd = [GIT, 'rev-parse', rev]
            return self._command(cmd).decode().rstrip()

    def ls(
        self, rev, path, recursive=False, recursive_dirs=False,
        directory=False, report=()
    ):
        path = type(self).cleanPath(path)
        forcedir = False
        if path.endswith('/'):
            forcedir = True
            path = path.rstrip('/')
        ltrim = len(path)

        # make sure the path exists
        if path == '':
            if directory:
                entry = attrdict(path='/', type='d')
                if 'commit' in report:
                    entry.commit = self.canonical_rev(rev)
                return [entry]
        else:
            epath = path.rstrip('/').encode(self.encoding)
            cmd = [GIT, 'ls-tree', '-z', rev, '--', epath]
            output = self._command(cmd)
            if not output:
                raise PathDoesNotExist(rev, path)
            meta, ename = output.split(b'\t', 1)
            meta = meta.decode().split()
            if meta[1] == 'tree':
                if not (directory or path.endswith('/')):
                    path = path + '/'
            elif forcedir:
                raise PathDoesNotExist(rev, path)

        cmd = [GIT, 'ls-tree', '-z']
        if recursive:
            cmd.append('-r')
            if recursive_dirs:
                cmd.append('-t')
        if 'size' in report:
            cmd.append('-l')
        epath = path.encode(self.encoding)
        cmd.extend([rev, '--', epath])
        output = self._command(cmd).rstrip(b'\0')
        if not output:
            return []

        results = []
        for line in output.split(b'\0'):
            meta, ename = line.split(b'\t', 1)
            meta = meta.decode().split()
            mode = int(meta[0], 8)
            objid = meta[2]
            name = ename.decode(self.encoding, 'replace')
            if recursive_dirs and path == name + '/':
                continue
            assert name.startswith(path), 'unexpected output: ' + str(line)
            entry = attrdict(path=name)
            entry_name = name[ltrim:].lstrip('/')
            if entry_name:
                entry.name = entry_name
            if stat.S_ISDIR(mode):
                entry.type = 'd'
            elif stat.S_ISREG(mode):
                entry.type = 'f'
                if 'executable' in report:
                    entry.executable = bool(mode & stat.S_IXUSR)
                if 'size' in report:
                    entry.size = int(meta[3])
            elif stat.S_ISLNK(mode):
                entry.type = 'l'
                if 'target' in report:
                    entry.target = self._cat(rev, ename).decode(self.encoding, 'replace')
            else:
                assert False, 'unexpected output: ' + str(line)
            if 'commit' in report:
                try:
                    entry.commit = self._object_cache[objid]
                    entry._commit_cached = True
                except KeyError:
                    ename = name.encode(self.encoding)
                    cmd = [GIT, 'log', '--pretty=format:%H', '-1', rev, '--', ename]
                    commit = self._command(cmd).decode()
                    entry.commit = self._object_cache[objid] = commit
            results.append(entry)

        return results

    def _cat(self, rev, path):
        rp = rev.encode('ascii') + b':' + path
        cmd = [GIT, 'cat-file', 'blob', rp]
        return self._command(cmd)

    def cat(self, rev, path):
        path = type(self).cleanPath(path)
        ls = self.ls(rev, path, directory=True)
        assert len(ls) == 1
        if ls[0].get('type') != 'f':
            raise BadFileType(rev, path)
        epath = path.encode(self.encoding, 'strict')
        return self._cat(rev, epath)

    def readlink(self, rev, path):
        path = type(self).cleanPath(path)
        ls = self.ls(rev, path, directory=True)
        assert len(ls) == 1
        if ls[0].get('type') != 'l':
            raise BadFileType(rev, path)
        epath = path.encode(self.encoding, 'strict')
        return self._cat(rev, epath).decode(self.encoding, 'replace')

    def branches(self):
        cmd = [GIT, 'branch']
        output = self._command(cmd).decode(self.encoding, 'replace')
        results = []
        for line in output.splitlines():
            m = branch_rx.match(line)
            assert m, 'unexpected output: ' + str(line)
            results.append(m.group('name'))
        return results

    def tags(self):
        cmd = [GIT, 'tag']
        output = self._command(cmd).decode(self.encoding, 'replace')
        return output.splitlines()

    def heads(self):
        return self.branches() + self.tags()

    def empty(self):
        cmd = [GIT, 'rev-parse', 'HEAD']
        p = subprocess.Popen(
            cmd, cwd=self.path, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        stdout, stderr = p.communicate()
        return not rev_rx.match(stdout.decode())

    def __contains__(self, rev):
        cmd = [GIT, 'rev-list', '-n', '1', rev]
        p = subprocess.Popen(
            cmd, cwd=self.path, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        stdout, stderr = p.communicate()
        return p.returncode == 0

    def __len__(self):
        cmd = [GIT, 'rev-list', '--all']
        p = subprocess.Popen(
            cmd, cwd=self.path, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        stdout, stderr = p.communicate()
        return len(stdout.splitlines())

    def log(
        self, revrange=None, limit=None, firstparent=False, merges=None,
        path=None, follow=False
    ):
        cmd = [GIT, 'log', '-z', '--pretty=format:%H%n%P%n%ai%n%an <%ae>%n%B', '--encoding=none']
        if limit is not None:
            cmd.append('-' + str(limit))
        if firstparent:
            cmd.append('--first-parent')
        if merges is not None:
            if merges:
                cmd.append('--merges')
            else:
                cmd.append('--no-merges')
        single = False
        if revrange is None:
            if self.empty():
                return []
            cmd.append('--all')
        elif isinstance(revrange, (tuple, list)):
            if revrange[0] is None:
                if revrange[1] is None:
                    if self.empty():
                        return []
                    cmd.append('--all')
                else:
                    cmd.append(revrange[1])
            else:
                if revrange[1] is None:
                    cmd.append(revrange[0] + '..')
                else:
                    cmd.append(revrange[0] + '..' + revrange[1])
        else:
            entry = self._commit_cache.get(self.canonical_rev(revrange))
            if entry:
                entry._cached = True
                return entry
            cmd.extend(['-1', revrange])
            single = True
        if path:
            if follow:
                cmd.append('--follow')
            cmd.extend(['--', type(self).cleanPath(path)])
        output = self._command(cmd).decode(self.encoding, 'replace')

        results = []
        for log in output.split('\0'):
            rev, parents, date, author, message = log.split('\n', 4)
            parents = parents.split()
            date = parse_isodate(date)
            entry = CommitLogEntry(rev, parents, date, author, message)
            if rev not in self._commit_cache:
                self._commit_cache[rev] = entry
            if single:
                return entry
            results.append(entry)
        return results

    def changed(self, rev):
        cmd = [GIT, 'diff-tree', '-z', '-C', '-r', '-m', '--first-parent', '--root', rev]
        output = self._command(cmd)
        results = []
        for line in output.rstrip(b'\0').split(b'\0:')[1:]:
            path = line.split(b'\0')
            meta = path.pop(0).split()
            status = meta[3].decode()[0]
            src_path = path[0].decode(self.encoding, 'replace')
            if len(path) == 2:
                dst_path = path[1].decode(self.encoding, 'replace')
                entry = FileChangeInfo(dst_path, str(status), src_path)
            else:
                entry = FileChangeInfo(src_path, str(status))
            results.append(entry)
        return results

    def pdiff(self, rev):
        cmd = [GIT, 'diff-tree', '-p', '-r', '-m', '--first-parent', '--root', rev]
        return self._command(cmd)

    def diff(self, rev_a, rev_b, path=None):
        cmd = [GIT, 'diff', rev_a, rev_b]
        if path is not None:
            cmd.extend(['--', type(self).cleanPath(path)])
        return self._command(cmd)

    def ancestor(self, rev1, rev2):
        cmd = [GIT, 'merge-base', rev1, rev2]
        p = subprocess.Popen(
            cmd, cwd=self.path, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        stdout, stderr = p.communicate()
        if p.returncode == 0:
            return stdout.decode().rstrip()
        elif p.returncode == 1:
            return None
        else:
            raise subprocess.CalledProcessError(p.returncode, cmd, stderr)

    def blame(self, rev, path):
        path = type(self).cleanPath(path)
        ls = self.ls(rev, path, directory=True)
        assert len(ls) == 1
        if ls[0].get('type') != 'f':
            raise BadFileType(rev, path)
        cmd = [GIT, 'blame', '--root', '--encoding=none', '-p', rev, '--', path]
        output = self._command(cmd)
        rev = None
        revinfo = {}
        results = []
        for line in output.splitlines():
            if line.startswith(b'\t'):
                ri = revinfo[rev]
                author = ri['author'] + ' ' + ri['author-mail']
                ts = int(ri['author-time'])
                tz = UTCOffset(str(ri['author-tz']))
                date = datetime.datetime.fromtimestamp(ts, tz)
                entry = BlameInfo(rev, author, date, line[1:])
                results.append(entry)
            else:
                k, v = line.decode(self.encoding, 'replace').split(None, 1)
                if rev_rx.match(k):
                    rev = k
                else:
                    revinfo.setdefault(rev, {})[k] = v
        return results
